from ._shell import shell
from ._gen import gen
from ._await import await_
from ._set import set
from ._block import block
from ._if import if_
from ._unless import unless
from ._add import add
from ._select import select
from ._each import each
from ._geneach import geneach
from ._strip import strip
from ._subtract import subtract
from ._role import role
from ._user import user
from ._system import system
from ._assistant import assistant
from ._break import break_
from ._equal import equal
from ._greater import greater
from ._less import less


import guidance
from ..utils import get_openai_llm, get_transformers_llm

def test_select():
    """ Test the behavior of `select`.
    """

    llm = get_openai_llm("text-curie-001")
    program = guidance("Is Everest very tall?\nAnswer 'Yes' or 'No': '{{#select 'name'}}Yes{{or}}No{{/select}}", llm=llm)
    out = program()
    assert out["name"] in ["Yes", "No"]

def test_select_longtext():
    """ Test the behavior of `select`.
    """

    llm = get_openai_llm("text-curie-001")
    program = guidance("""Is Everest very tall?\nAnswer:
{{#select 'name'}}No because of all the other ones.{{or}}Yes because I saw it.{{/select}}""", llm=llm)
    out = program()
    assert out["name"] in ["No because of all the other ones.", "Yes because I saw it."]

def test_select_longtext_transformers():
    """ Test the behavior of `select`.
    """

    llm = get_transformers_llm("gpt2")
    program = guidance("""Is Everest very tall?\nAnswer:
{{#select 'name'}}No because of all the other ones.{{or}}Yes because I saw it.{{/select}}""", llm=llm)
    out = program()
    assert out["name"] in ["No because of all the other ones.", "Yes because I saw it."]

def test_select_with_list():
    """ Test the behavior of `select` in non-block mode.
    """

    llm = get_openai_llm("text-curie-001")
    program = guidance("Is Everest very tall?\nAnswer 'Yes' or 'No': '{{select 'name' options=options}}", llm=llm)
    out = program(options=["Yes", "No"])
    assert out["name"] in ["Yes", "No"]

def test_select_list_append():
    """ Test the behavior of `select` with list_append=True.
    """

    llm = get_openai_llm("text-curie-001")
    program = guidance("Is Everest very tall?\n{{select 'name' options=options list_append=True}}\n{{select 'name' options=options list_append=True}}", llm=llm)
    out = program(options=["Yes", "No"])
    assert len(out["name"]) == 2
    for v in out["name"]:
        assert v in ["Yes", "No"]

def test_select_odd_spacing():
    """ Test the behavior of `select` with list_append=True.
    """

    llm = get_openai_llm("text-curie-001")
    prompt = guidance('''Is the following sentence offensive? Please answer with a single word, either "Yes", "No", or "Maybe".
    Sentence: {{example}}
    Answer: {{#select "answer" logprobs='logprobs'}} Yes{{or}} Nein{{or}} Maybe{{/select}}''', llm=llm)
    prompt = prompt(example='I hate tacos.')
    assert prompt["answer"] in [" Yes", " Nein", " Maybe"]


def test_overlapping_options():
    """ Test the behavior of `select` when one option is a prefix of another.
    """

    llm = get_transformers_llm("gpt2")
    options = ['a', 'aa']
    program = guidance("'{{select options=options}}", llm=llm)
    out = program(options=options)
    assert out["selected"] in options

# TODO: fix this next
# def test_unexpected_tokens():
#     """ Test the behavior of `select` when the next tokens are hard to predict.
#     """

#     llm = get_transformers_llm("gpt2")
#     options = ['a', 'b']
#     program = guidance("some word xy{{select options=options}}", llm=llm)
#     out = program(options=options)
#     assert out["selected"] in options

# TODO: test when we have few starting tokens


from ._role import role

async def system(hidden=False, _parser_context=None):
    ''' A chat role block for the 'system' role.

    This is just a shorthand for {{#role 'system'}}...{{/role}}.

    Parameters
    ----------
    hidden : bool
        Whether to include the assistant block in future LLM context. 
    '''
    return await role(name="system", hidden=hidden, _parser_context=_parser_context)
system.is_block = True


import guidance

def test_equal():
    """ Test the behavior of `equal`.
    """

    program = guidance("""{{#if (equal val 5)}}are equal{{else}}not equal{{/if}}""")
    assert str(program(val=4)) == "not equal"
    assert str(program(val=5)) == "are equal"
    assert str(program(val="5")) == "not equal"

def test_equal_with_symbol():
    """ Test the behavior of `equal` written as `==`.
    """

    program = guidance("""{{#if (== val 5)}}are equal{{else}}not equal{{/if}}""")
    assert str(program(val=4)) == "not equal"
    assert str(program(val=5)) == "are equal"
    assert str(program(val="5")) == "not equal"


import inspect
import re
import uuid
from .._utils import strip_markers

async def each(list, hidden=False, filter=None, _parser_context=None):
    ''' Iterate over a list and execute a block for each item.

    Parameters
    ----------
    list : iterable
        The list to iterate over. Inside the block each element will be available as `this`.
    hidden : bool
        Whether to include the generated item blocks in future LLM context.
    filter : callable
        A function to filter the list before iterating over it. The function should take a single
        argument (the list) and return a new list. It can also optionally have a `template_context`
        argument that will be set to a dictionary containing the following keys:
            - "@block_text": the text of the block
            - "@tokenizer": the tokenizer used to parse the block
    '''
    block_content = _parser_context['block_content']
    parser = _parser_context['parser']
    parser_node = _parser_context['parser_node']
    assert len(block_content) == 1
    echo = not hidden

    if not parser.executing:
        return parser_node.text

    # apply a filter to the list if one was provided
    if filter is not None:

        # if the list is callable then we call it to get an iterable
        if callable(filter):
            # we support a set of optional arguments to the list function
            named_args = {}
            sig = inspect.signature(filter)
            if "template_context" in sig.parameters:
                named_args["template_context"] = {
                    "@block_text": block_content[0].text,
                    "@tokenizer": parser.program.tokenizer
                }
            list = filter(list, **named_args)
        else:
            raise TypeError("Can't apply a non-callable filter: " + str(filter))

    # make sure the list is iterable
    try:
        iter(list)
    except TypeError:
        raise TypeError("The #each command cannot iterate over a non-iterable value: " + str(list))

    out = []
    partial_out = ""
    
    parser.variable_stack.append({})
    for i, item in enumerate(list):
        parser.variable_stack[-1]["@index"] = i
        parser.variable_stack[-1]["@first"] = i == 0
        parser.variable_stack[-1]["@last"] = i == len(list) - 1
        parser.variable_stack[-1]["this"] = item
        item_out = await parser.visit(
            block_content[0],
            next_node=_parser_context["next_node"],
            next_next_node=_parser_context["next_next_node"],
            prev_node=_parser_context["prev_node"]
        )
        if not echo:
            parser._trim_prefix(item_out)
        out.append(item_out)

        # check if the block has thrown a stop iteration signal
        if parser.caught_stop_iteration:
            parser.caught_stop_iteration = False
            break
    parser.variable_stack.pop()
    
    suffix = ""
    if not parser.executing:
        if isinstance(list, str):
            suffix = partial_out + parser_node.text
    
    if echo:
        return "{{!--GMARKER_each$$--}}" + "{{!--GMARKER_each$$--}}".join(out) + "{{!--GMARKER_each$$--}}" + suffix
    else:
        id = uuid.uuid4().hex
        l = len(out)
        out_str = "{{!--" + f"GMARKER_each_noecho_start_{echo}_{l}${id}$" + "--}}"
        for i, value in enumerate(out):
            if i > 0:
                out_str += "{{!--" + f"GMARKER_each_noecho_{echo}_{i}${id}$" + "--}}"
            out_str += value
        return out_str + "{{!--" + f"GMARKER_each_noecho_end${id}$" + "--}}"

        # return "{{!--GMARKER_each_noecho$$}}" + "{{!--GMARKER_each_noecho$$}}".join(out) + "{{!--GMARKER_each_noecho$$}}"
each.is_block = True


import guidance

def test_each():
    """ Test an each loop.
    """

    prompt = guidance("Hello, {{name}}!{{#each names}} {{this}}{{/each}}")
    assert str(prompt(name="Guidance", names=["Bob", "Sue"])) == "Hello, Guidance! Bob Sue"

def test_each_with_objects():
    """ Test an each loop with objects.
    """

    prompt = guidance("Hello, {{name}}!{{#each names}} {{this.name}}{{/each}}")
    out = prompt(
        name="Guidance",
        names=[{"name": "Bob"}, {"name": "Sue"}]
    )
    assert str(out) == "Hello, Guidance! Bob Sue"

def test_missing_list():
    prompt = guidance('''List of ideas:{{#each ideas}}test{{this}}{{/each}}''', await_missing=True)
    assert str(prompt()) == "List of ideas:{{#each ideas}}test{{this}}{{/each}}"
    # try:
    #     out = prompt()
    # except KeyError:
    #     return
    # assert False, "An error should have been raised because the list is missing!"


import guidance

def test_user():
    """ Basic test of `user`.
    """

    llm = guidance.llms.Mock("the output")

    program = guidance("""
{{~#system}}You are fake.{{/system}}
{{#user}}You are real.{{/user}}
{{#assistant}}{{gen 'output' save_prompt='prompt'}}{{/assistant}}""", llm=llm)
    out = program()
    assert '<|im_start|>user\nYou are real.<|im_end|>' in str(out)


import guidance
import pytest
from ..utils import get_llm

def test_gen():
    """ Test that LM generation works.
    """

    llm = guidance.llms.Mock(" Sue")
    prompt = guidance("Hello my name is{{gen 'name' max_tokens=5}}", llm=llm)
    out = prompt()
    assert len(out["name"]) > 1

def test_gen_n_greater_than_one():
    llm = guidance.llms.Mock(["mock output 0", "mock output 1", "mock output 2"])
    prompt = guidance('''The best thing about the beach is{{gen 'best' n=3 temperature=0.7 max_tokens=5}}''', llm=llm)
    a = prompt()
    assert "\n".join(a["best"]) == 'mock output 0\nmock output 1\nmock output 2'

def test_gen_n_greater_than_one_hidden():
    llm = guidance.llms.Mock()

    def aggregate(best):
        return '\n'.join(['- ' + x for x in best])
    prompt = guidance('''The best thing about the beach is{{gen 'best' temperature=0.7 n=3 hidden=True}}
{{aggregate best}}''', llm=llm)
    a = prompt(aggregate=aggregate)
    assert str(a) == 'The best thing about the beach is\n- mock output 0\n- mock output 1\n- mock output 2'

def test_pattern():
    import re
    llm = get_llm("transformers:gpt2")
    out = guidance('''On a scale of 1-10 I would say it is: {{gen 'score' pattern="[0-9]+"}}''', llm=llm)()
    assert re.match(r'[0-9]+', out["score"])

def test_pattern2():
    import re

    prompt = '''
    Tweak this proverb to apply to machine learning model instructions instead.

    {{proverb}}
    - {{book}} {{chapter}}:{{verse}}

    UPDATED
    Where there is no guidance, a people falls, but in an abundance of counselors there is safety.
    - GPT {{gen 'chapter' pattern='[0-9]' max_tokens=1}}:{{gen 'verse' pattern='[0-9]+' stop='\\n'}}
    '''[1:-1]

    llm = get_llm("transformers:gpt2")
    program = guidance(prompt, llm=llm)
    executed_program = program(
        proverb="Where there is no guidance, a people falls,\nbut in an abundance of counselors there is safety.",
        book="Proverbs",
        chapter=11,
        verse=14
    )

    assert re.fullmatch(r"[0-9]", executed_program["chapter"])
    assert re.fullmatch(r"[0-9]+", executed_program["verse"])

@pytest.mark.parametrize("llm", ["transformers:gpt2", "openai:text-curie-001"])
def test_stop(llm):
    """ Test that the stop argument works as expected.
    """
    llm = get_llm(llm)
    program = guidance("""Write "repeat this. " 10 times: repeat this. repeat this. repeat this. repeat this. repeat this. repeat this.{{gen stop="this" max_tokens=10}}""", llm=llm)
    out = program()
    assert str(out) == "Write \"repeat this. \" 10 times: repeat this. repeat this. repeat this. repeat this. repeat this. repeat this. repeat "

@pytest.mark.parametrize("llm", ["transformers:gpt2", "openai:text-curie-001"])
def test_stop_regex(llm):
    """ Test that the stop_regex argument works as expected.
    """
    llm = get_llm(llm)
    program = guidance("""Write "repeat this. " 10 times: repeat this. repeat this. repeat this. repeat this. repeat this. repeat this.{{gen stop_regex="th.s" max_tokens=10}}""", llm=llm)
    out = program()
    assert str(out) == "Write \"repeat this. \" 10 times: repeat this. repeat this. repeat this. repeat this. repeat this. repeat this. repeat "


from ._role import role

async def assistant(hidden=False, _parser_context=None):
    ''' A chat role block for the 'assistant' role.

    This is just a shorthand for {{#role 'assistant'}}...{{/role}}.

    Parameters
    ----------
    hidden : bool
        Whether to include the assistant block in future LLM context. 
    '''
    return await role(name="assistant", hidden=hidden, _parser_context=_parser_context)
assistant.is_block = True


def subtract(minuend, subtrahend):
    ''' Subtract the second variable from the first.

    Parameters
    ----------
    minuend : int or float
        The number to subtract from.
    subtrahend : int or float
        The number to subtract.
    '''
    return minuend - subtrahend


import itertools
import pygtrie
import numpy as np

async def select(variable_name="selected", options=None, logprobs=None, list_append=False, _parser_context=None):
    ''' Select a value from a list of choices.

    Parameters
    ----------
    variable_name : str
        The name of the variable to set with the selected value.
    options : list of str or None
        An optional list of options to select from. This argument is only used when select is used in non-block mode.
    logprobs : str or None
        An optional variable name to set with the logprobs for each option. If this is set the log probs of every option
        is fully evaluated. When this is None (the default) we use a greedy max approach to select the option (similar to
        how greedy decoding works in a language model). So in some cases the selected option can change when logprobs is
        set since it will be more like an exhaustive beam search scoring than a greedy max scoring.
    list_append : bool
        Whether to append the generated value to a list stored in the variable. If set to True, the variable
        must be a list, and the generated value will be appended to the list.
    '''
    parser = _parser_context['parser']
    block_content = _parser_context['block_content']
    parser_prefix = _parser_context['parser_prefix']
    partial_output = _parser_context['partial_output']

    if block_content is None:
        assert options is not None, "You must provide an options list like: {{select 'variable_name' options}} when using the select command in non-block mode."
    else:
        assert len(block_content) > 1, "You must provide at least one two options to the select block command."
        assert options is None, "You cannot provide an options list when using the select command in block mode."

    if options is None:
        options = [block_content[0].text]
        for i in range(1, len(block_content), 2):
            assert block_content[i].text == "{{or}}"
            options.append(block_content[i+1].text)

    # build a trie of the options
    token_map = pygtrie.CharTrie()
    for i,option in enumerate(options):
        token_map[option] = i
    
    async def recursive_select(current_prefix, allow_token_extension=True):
        """ This returns a dictionary of scores for each option (keyed by the option index).
        """

        # find which select options are possible
        try:
            extension_options = token_map.items(prefix=current_prefix)
        except KeyError:
            return {}

        # this is the dictionary of logprobs for each option we will return
        # note that the logprobs are just for this branch point and below in the decision tree
        logprobs_out = {option[0]: -1000 for option in extension_options}

        # extend the prefix with the longest common prefix among the valid options
        # we also stop early if we have one option
        if len(extension_options) == 1:
            logprobs_out[extension_options[0][0]] = 0 # probability of 1.0 that we will select the only valid option
            return logprobs_out
        else:
            match_index = 0
            for i in range(len(current_prefix), min([len(o[0]) for o in extension_options])):
                if len(set([o[0][i] for o in extension_options])) > 1:
                    match_index = i
                    break
            if match_index > len(current_prefix):
                current_prefix += extension_options[0][0][len(current_prefix):match_index]
                # extension_options = [(option[i:], index) for option,index in extension_options]

        # bias the logits towards valid options
        # TODO: this retokenizes the whole prefix many times, perhaps this could become a bottleneck?
        tmp_prefix = parser_prefix + current_prefix # this is so we get the right tokenization at the boundary
        tmp_prefix_tokens = parser.program.llm.encode(tmp_prefix)
        logit_bias1 = {} # token extension biases
        logit_bias2 = {} # next token biases
        for option,index in extension_options:
            option_tokens = parser.program.llm.encode(parser_prefix + option)
            
            # if we extended the last token to a longer one
            if len(tmp_prefix_tokens) > 0 and option_tokens[len(tmp_prefix_tokens)-1] != tmp_prefix_tokens[-1]:
                if allow_token_extension: # this is a valid extension only if we are not allowed to extend the token
                    logit_bias1[option_tokens[len(tmp_prefix_tokens)-1]] = 100
            
            # if we did not extend the last token to a longer one we can bias the next token
            else:
                if len(option_tokens) > len(tmp_prefix_tokens):
                    logit_bias2[option_tokens[len(tmp_prefix_tokens)]] = 100

            # logit_bias[option_tokens[len(tmp_prefix_tokens)-1]] = tmp_prefix_tokens[-1]
            # if len(option_tokens) > len(tmp_prefix_tokens) and :
            #     logit_bias[option_tokens[len(tmp_prefix_tokens)]] = 100

        # extend the prefix by extending the last token
        if len(logit_bias1) > 0:
            call_prefix = parser.program.llm.decode(tmp_prefix_tokens[:-1])
            logit_bias = logit_bias1
            last_token_str = parser.program.llm.decode(tmp_prefix_tokens[-1:])

        # extend the prefix by adding the next token
        else:
            call_prefix = tmp_prefix
            logit_bias = logit_bias2
            last_token_str = ""

        # generate the token logprobs
        gen_obj = await parser.llm_session(
            call_prefix, # TODO: perhaps we should allow passing of token ids directly? (this could allow us to avoid retokenizing the whole prefix many times)
            max_tokens=1,
            logit_bias=logit_bias,
            logprobs=10,
            cache_seed=0,
            token_healing=False # we manage token boundary healing ourselves for this function
        )
        logprobs_result = gen_obj["choices"][0]["logprobs"]
        top_logprobs = logprobs_result["top_logprobs"][0]
        
        # no need to explore all branches if we are just taking the greedy max
        if logprobs is None:
            max_key = max(top_logprobs, key=top_logprobs.get)
            top_logprobs = {max_key: top_logprobs[max_key]}

        # for each possible next token, see if it grows the prefix in a valid way
        for token_str,logprob in top_logprobs.items():

            # build our recursive call prefix
            if len(last_token_str) > 0:
                rec_prefix = current_prefix[:-len(last_token_str)]
            else:
                rec_prefix = current_prefix
            rec_prefix += token_str
            
            # if we did not extend the last token then we recurse while ignoring the possibility of extending the last token
            if token_str == last_token_str:
                sub_logprobs = await recursive_select(rec_prefix, allow_token_extension=False)
            else:
                sub_logprobs = await recursive_select(rec_prefix)
            
            # we add the logprob of this token to the logprob of the suffix
            for k in sub_logprobs:
                logprobs_out[k] = sub_logprobs[k] + logprob

        # if we did token healing and did not extend past our prefix we need to consider the next token
        # TODO: when returning all logprobs we need to consider all the options, which means we should
        # force the model to not token heal and see what would have happened then on the next token...
        # first_token_str = max(top_logprobs, key=top_logprobs.get)
        # if len(logprobs_result["top_logprobs"]) > 1 and len(first_token_str) == remove_prefix:            
        #     top_logprobs = logprobs_result["top_logprobs"][1]
        #     for token_str,logprob in top_logprobs.items():
        #         sub_logprobs = await recursive_select(current_prefix + token_str)
        #         for k in sub_logprobs:

        #             # compute the probability of a logical OR between the new extension and the previous possible ones
        #             p1 = np.exp(logprobs_out[k])
        #             p2 = np.exp(sub_logprobs[k] + logprob)
        #             or_prob = p1 + p2 - p1*p2
        #             logprobs_out[k] = np.log(or_prob)

        return logprobs_out
        
    # recursively compute the logprobs for each option
    option_logprobs = await recursive_select("")

    selected_option = max(option_logprobs, key=option_logprobs.get)
    
    # see if we are appending to a list or not
    if list_append:
        value_list = parser.get_variable(variable_name, [])
        value_list.append(selected_option)
        parser.set_variable(variable_name, value_list)
        if logprobs is not None:
            logprobs_list = parser.get_variable(logprobs, [])
            logprobs_list.append(option_logprobs)
            parser.set_variable(logprobs, logprobs_list)
    else:
        parser.set_variable(variable_name, selected_option)
        if logprobs is not None:
            parser.set_variable(logprobs, option_logprobs)
    
    if max(option_logprobs.values()) <= -1000:
        raise ValueError("No valid option generated in #select! Please post a GitHub issue since this should not happen :)")
    
    partial_output(selected_option)

select.is_block = True


import guidance

def test_if():
    """ Test the behavior of `if`.
    """

    prompt = guidance("""Answer: {{#if flag}}Yes{{/if}}""")

    for flag in [True, 1, "random text"]:
        out = prompt(flag=flag)
        assert str(out) == "Answer: Yes"
    
    for flag in [False, 0, ""]:
        out = prompt(flag=flag)
        assert str(out) == "Answer: "

def test_if_else():
    """ Test the behavior of `if` with an `else` clause.
    """

    prompt = guidance("""Answer 'Yes' or 'No': '{{#if flag}}Yes{{else}}No{{/if}}'""")

    for flag in [True, 1, "random text"]:
        out = prompt(flag=flag)
        assert str(out) == "Answer 'Yes' or 'No': 'Yes'"
    
    for flag in [False, 0, ""]:
        out = prompt(flag=flag)
        assert str(out) == "Answer 'Yes' or 'No': 'No'"


import guidance
from ..utils import get_transformers_llm

def test_basic():
    llm = get_transformers_llm('gpt2')
    with llm.session() as s:
        out = s("this is a test", max_tokens=5)
        print(out)

def test_repeat():
    llm = get_transformers_llm('gpt2')
    with llm.session() as s:
        out1 = s("this is a test", max_tokens=5)
        out2 = s("this is a test like another", max_tokens=5)
        print(out2)



def test_select():
    llm = get_transformers_llm('gpt2')
    program = guidance('''Answer "yes" or "no": "{{#select 'answer'}}yes{{or}}no{{/select}}"''', llm=llm)
    out = program()
    assert out["answer"] in ["yes", "no"]



import os
import time
import collections
import regex
import pygtrie
import queue
import threading
import logging
from .._llm import LLM, LLMSession, SyncSession
from .._transformers import Transformers, TransformersSession

class StableLMChat(Transformers):
    """ A HuggingFace transformers version of the StableLM language model with Guidance support.
    """

    cache = LLM._open_cache("_stablelm.diskcache")

    @staticmethod
    def role_start(role):
        return "<|"+role.upper()+"|>"
    
    @staticmethod
    def role_end(role):
        return '' 